#!/usr/bin/env python

# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/12/27 14:42
# @Last Modified by: wqshen

import os
import sys
import argparse
import logzero
import pandas as pd
import xarray as xr
from logzero import logger
from datetime import datetime
from pymdfs import MdfsClient


def construct_slice(coord):
    """construct point, slice, or step slice from list or tuple

    Parameters
    ----------
    coord : list or tuple
        (point,) or (start, stop) or (start, stop, step)

    Returns
    -------
    s : float, slice
        slice or point
    """
    if len(coord) == 1:
        s = coord[0]
    elif 1 < len(coord) < 3:
        s = slice(*coord)
    else:
        raise Exception("length of coord must less than or equal 3")
    return s


def tpyecast(string):
    """cast string into int/float/str"""
    try:
        return int(string)
    except ValueError:
        try:
            return float(string)
        except Exception:
            return string


def _main():
    example_text = """Example:
     client_dump ECMWF_HR 2023021920 -f 24 --level 500 -v RH,UGRD,VGRD,TMP,HGT -e ECMWF_HR.2023021920.nc
     """

    args_parser = lambda s: [tpyecast(item) for item in s.split(',')]
    parser = argparse.ArgumentParser(description='MDFS Data Dumper', epilog=example_text,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('datasource', help='data source name')
    parser.add_argument('inittime', help='model initial datetime or observation datetime')
    parser.add_argument('-f', '--fh', help='model forecast hour', type=args_parser)
    parser.add_argument('-e', '--outfile', type=str, help='output netcdf file name')
    parser.add_argument('-c', '--complevel', type=str, help='output netcdf4 compress level',
                        default=5)
    parser.add_argument('-v', '--varname', help='model variable names', type=args_parser)
    parser.add_argument('-x', '--lon', help='longitude point or range', type=args_parser)
    parser.add_argument('-y', '--lat', help='latitude point or range', type=args_parser)
    parser.add_argument('-p', '--level', help='pressure level point or range', type=args_parser)
    parser.add_argument('-t', '--offset-inittime', help='offset inittime (hours) to variable',
                        type=str)
    parser.add_argument('--name_map', help='map variable name to new', type=args_parser)
    parser.add_argument('-s', '--server', type=str, help='GDS server address',
                        default='xxx.xxx.xxx.xxx')
    parser.add_argument('-o', '--loglevel', type=int, help='logger level in number', default=20)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    logzero.loglevel(args.loglevel)
    logger.debug(args)

    inittime = datetime.strptime(args.inittime, '%Y%m%d%H')
    coord_kwargs = dict()
    if args.lon is not None:
        coord_kwargs['lon'] = construct_slice(args.lon)
    if args.lat is not None:
        coord_kwargs['lat'] = construct_slice(args.lat)
    if args.level is not None:
        coord_kwargs['level'] = construct_slice(args.level)

    with MdfsClient(args.server) as mc:
        dataset = mc.sel(args.datasource, inittime, args.fh, args.varname, **coord_kwargs)
        logger.debug(dataset)
        if isinstance(dataset, list):
            if isinstance(dataset[0], xr.DataArray):
                dataset = xr.merge(dataset)
            elif isinstance(dataset[0], pd.DataFrame):
                dataset = pd.concat(dataset)
        elif isinstance(dataset, xr.DataArray):
            dataset = dataset.to_dataset()
        else:
            raise TypeError(f"type of return ({type(dataset)}) not be understood.")

        if args.offset_inittime is not None:
            dataset['time'] = pd.to_datetime(dataset.time.values) + pd.to_timedelta(
                args.offset_inittime)

        if args.name_map is not None:
            dataset = dataset.rename({args.name_map[0]: args.name_map[1]})

    logger.info(dataset)
    if args.outfile is not None:
        _, file_extension = os.path.splitext(args.outfile)
        if file_extension in ('.nc', '.nc4') and isinstance(dataset, xr.Dataset):
            comp = dict(zlib=True, complevel=args.complevel)
            encoding = {var: comp for var in dataset.data_vars}
            dataset.to_netcdf(args.outfile, encoding=encoding)
        elif file_extension in ('.txt', '.csv'):
            if isinstance(dataset, xr.DataArray):
                dataset.to_dataframe().to_csv(args.outfile)
            else:
                dataset.to_csv(args.outfile)
        else:
            raise NotImplementedError("Only support nc, nc4, txt, csv extentsion.")


if __name__ == '__main__':
    _main()



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2023/2/9 22:07
# @Last Modified by: wqshen

import numpy as np
import xarray as xr
from typing import Optional, Union, cast
from struct import unpack
from datetime import datetime, timedelta
from dataclasses import dataclass, fields


@dataclass
class MdfsGridHead:
    discriminator: str
    dtype: int
    modelName: str
    element: str
    description: str
    level: float
    year: int
    month: int
    day: int
    hour: int
    timezone: int
    period: int
    startLongitude: float
    endLongitude: float
    longitudeGridSpace: float
    latitudeGridNumber: int
    startLatitude: float
    endLatitude: float
    latitudeGridSpace: float
    longitudeGridNumber: int
    isolineStartValue: float
    isolineEndValue: float
    isolineSpace: float
    Extent: str

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            if isinstance(field_value, bytes):
                field_value = field_value.decode('GBK').rstrip('\x00')
            setattr(self, field.name, field.type(field_value))


@dataclass
class MdfsGridData(object):
    head: Optional[MdfsGridHead] = None
    data: Optional[Union[np.ndarray, dict]] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[Union[xr.DataArray, xr.Dataset]] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, **kwargs):
        """interface to call xarray.DataArray/xarray.Dataset sel method

        Parameters
        ----------
        kwargs (dict): other parameters used to process xarray.DataArray/xarray.Dataset variable

        Returns
        -------
        dar (xarray.DataArray, xarray.Dataset): sel variable
        """
        if self._ds.indexes['lat'].is_monotonic_decreasing:
            if 'lat' in kwargs and isinstance(kwargs.get('lat'), slice):
                lat = kwargs.get('lat')
                kwargs['lat'] = slice(lat.stop, lat.start, lat.step)
        dar = self._ds.sel(**kwargs)
        return dar

    def read(self, path_or_bytes: Union[str, bytes]):
        if isinstance(path_or_bytes, str):
            with open(path_or_bytes, 'rb') as f:
                bytes_array = f.read()
        else:
            bytes_array = path_or_bytes
        head_data = unpack('=4sh20s50s30sfiiiiiifffifffifff100s', bytes_array[:278])
        self.head = MdfsGridHead(*head_data)
        n_points = self.head.latitudeGridNumber * self.head.longitudeGridNumber
        body_data = np.frombuffer(bytes_array[278:278 + n_points * 4], '{}f'.format(n_points))
        data = np.array(body_data).reshape(self.head.longitudeGridNumber,
                                           self.head.latitudeGridNumber)
        if self.head.dtype == 11:
            self.data = {'magnitude': data}
            p = 278 + n_points * 4
            body_data_angle = unpack('{}f'.format(n_points), bytes_array[p:p + n_points * 4])
            self.data['angle'] = np.array(body_data_angle).reshape(self.head.longitudeGridNumber,
                                                                   self.head.latitudeGridNumber)
        else:
            self.data = data
        self._ds = self.to_xarray(name=self.head.element)

    def to_xarray(self, name=None):
        def set_mask_and_attrs(data):
            # To mask potential missing values
            if self.missing_value is not None:
                data = data.where(data == self.missing_value)

            # Append extra attributes to netcdf varibale
            data.attrs['model'] = self.head.modelName
            data.attrs['description'] = self.head.description
            data.attrs['inittime'] = inittime
            data.attrs['fh'] = fh

            if self.pathfile is not None:
                data.attrs['raw'] = self.pathfile
            return data

        head = self.head
        lons = np.linspace(head.startLongitude, head.endLongitude, head.latitudeGridNumber)
        lats = np.linspace(head.startLatitude, head.endLatitude, head.longitudeGridNumber)

        inittime = datetime(head.year, head.month, head.day, head.hour)
        fh = self.head.period
        leadtime = inittime + timedelta(hours=fh)
        dims = ('time', 'lat', 'lon')
        coords = {'time': [leadtime], 'lat': lats, 'lon': lons}
        if isinstance(self.data, dict):
            data_u = self.data['magnitude'] * np.cos(np.deg2rad(self.data['angle']))
            data_v = self.data['magnitude'] * np.sin(np.deg2rad(self.data['angle']))
            data_u = xr.DataArray(data_u[None, ...], dims=dims, coords=coords, name=name + '_U')
            data_v = xr.DataArray(data_v[None, ...], dims=dims, coords=coords, name=name + '_V')
            data = xr.merge([set_mask_and_attrs(data_u), set_mask_and_attrs(data_v)])
        else:
            data = xr.DataArray(self.data[None, ...], dims=dims, coords=coords, name=name)
            data = set_mask_and_attrs(data)

        return data

    def to_diamond(self, pathfile: str):
        from ..diamond.diamond4 import Diamond4, Diamond4Head

        head = self.head
        m4 = Diamond4()
        m4.head = Diamond4Head('diamond', head.dtype,
                               head.description, head.year, head.month,
                               head.day, head.hour, head.period, head.level,
                               head.longitudeGridSpace, head.latitudeGridSpace,
                               head.startLongitude, head.endLongitude,
                               head.startLatitude, head.endLatitude,
                               head.latitudeGridNumber, head.longitudeGridNumber,
                               20, -300, 300, 1, 0)
        m4.data = self._ds.squeeze().values
        m4.write(pathfile, fmt='%.2f')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close open dataset"""
        del self.pathfile, self.data, self.head, self._ds

    def __repr__(self) -> str:
        """print"""
        return self.head.__repr__()



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2021/2/3 17:26
# @Last Modified by: wqshen

import os
import numpy as np
import pandas as pd
import xarray as xr
from typing import Optional, Union, cast
from datetime import datetime, timedelta
from dataclasses import dataclass, fields, astuple


@dataclass
class Diamond4Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    duration: int
    level: float
    xinterval: float
    yinterval: float
    startlon: float
    endlon: float
    startlat: float
    endlat: float
    xsize: int
    ysize: int
    lineinteravl: float
    startvalue: float
    endvalue: float
    smooth: float = 1
    boldvalue: float = 0

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond4:
    """Micaps Diamond Data class

    From this class, one can read and write Micaps Diamond4 file

    Read Diamond4
    --------------
    >>> m4 = Diamond4(pathfile='../../sample_data/diamond_4/21032920.024')
    >>> m4.read()
    >>> print(m4)
    >>> print(m4.data.shape)
    >>> print(m4.to_xarray())

    Write Diamond4
    --------------
    >>> m4 = Diamond4()
    >>> m4.head = Diamond4Head('diamond', 4, '20年11月27日T63_500hPa温度120小时预报', 2020, 11, 27, 20, 120, 500,
    ...                         1.875, -1.875, 0, 180, 90, 0, 97, 49, 20, -300, 300, 1, 0)
    >>> m4.data = np.random.randint(-50, 10, size=(97, 49))
    >>> m4.write("20201127.120", fmt='%.2f')
    """
    head: Optional[Diamond4Head] = None
    data: Optional[np.ndarray] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[xr.DataArray] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, **kwargs):
        """interface to call `xarray.DataArray.sel`_ method

        Parameters
        ----------
        kwargs (dict): other parameters used to process xarray.DataArray variable

        Returns
        -------
        dar (xarray.DataArray): sel variable
        """
        if self._ds.indexes['lat'].is_monotonic_decreasing:
            if 'lat' in kwargs and isinstance(kwargs.get('lat'), slice):
                lat = kwargs.get('lat')
                kwargs['lat'] = slice(lat.stop, lat.start, lat.step)
        dar = self._ds.sel(**kwargs)
        return dar

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 4 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()
        headsize = len(fields(Diamond4Head))
        self.head = Diamond4Head(*data[:headsize])
        self.data = np.asfarray(data[headsize:]).reshape((self.head.ysize, self.head.xsize))
        self._ds = self.to_xarray(name='var')

    def write(self, pathfile: str, encoding: str = 'GBK', fmt: str = '%.2f'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str], default: '.5f'
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head[:3]) + "\n")
            f.write("\t".join(head[3:9]) + "\n")
            f.write("\t".join(head[9:22]) + "\n")

            np.savetxt(f, self.data, fmt=fmt, delimiter='\t')

    def to_netcdf(self, pathfile: str, name: str, complevel: int = 5):
        """write data into NetCDF4 file

        Parameters
        ----------
        pathfile : str
            path to file name
        name : str
            variable name when pack data into NetCDF4 Variable
        complevel : Optional[int], default: 5
            Compress level when pack data into NetCDF4 dataset
        """
        data = self.to_xarray()
        # convert to dataset and save to netcdf file with encoding
        comp = dict(zlib=True, complevel=complevel)
        ds = data.to_dataset(name=name)
        encoding = {var: comp for var in ds.data_vars}
        ds.to_netcdf(pathfile, encoding=encoding)

    def from_xarray(self, darray: xr.DataArray):
        """convert xarray into Diamond4"""
        desc = getattr(darray, 'long_name', 'Unknown')
        time = getattr(darray, 'time', datetime(1900, 1, 1, 0))
        it = pd.to_datetime(time)
        duration = getattr(darray, 'step', 0)
        level = getattr(darray, 'level', 0)
        xinterval = darray.lon[1] - darray.lon[0]
        yinterval = darray.lat[1] - darray.lat[0]
        startlon = darray.lon.min()
        startlat = darray.lat.min()
        endlon = darray.lon.max()
        endlat = darray.lat.max()
        xsize, ysize = darray.shape[-1], darray.shape[-2]
        startvalue = darray.min().floor()
        endvalue = darray.max().ceil()
        lineinteravl = (endvalue - startvalue) / 20.
        self.head = Diamond4Head('diamond', 4, desc, it.year, it.month, it.day, it.hour, duration,
                                 level, xinterval, yinterval, startlon, endlon, startlat, endlat,
                                 xsize, ysize, lineinteravl, startvalue, endvalue, )
        self.data = darray.values
        self._ds = darray

    def to_xarray(self, name=None):
        """Convert instance into xarray.DataArray

        Returns
        -------
        data: xr.DataArray
            converted data
        """
        lons = np.linspace(self.head.startlon, self.head.endlon, self.head.xsize)
        lats = np.linspace(self.head.startlat, self.head.endlat, self.head.ysize)

        if self.head.year < 50:
            inittime = datetime(2000 + self.head.year, self.head.month,
                                self.head.day, self.head.hour)
        elif self.head.year < 100:
            inittime = datetime(1900 + self.head.year, self.head.month,
                                self.head.day, self.head.hour)
        else:
            inittime = datetime(self.head.year, self.head.month,
                                self.head.day, self.head.hour)
        fh = self.head.duration
        leadtime = inittime + timedelta(hours=fh)
        data = xr.DataArray(self.data[None, ...], dims=('time', 'lat', 'lon'),
                            coords={'time': [leadtime], 'lat': lats, 'lon': lons},
                            name=name)
        # To mask potential missing values
        if self.missing_value is not None:
            data = data.where(data == self.missing_value)
        # Append extra attributes to netcdf varibale
        data.attrs['description'] = self.head.description
        data.attrs['inittime'] = inittime
        data.attrs['fh'] = fh

        if self.pathfile is not None:
            data.attrs['raw'] = self.pathfile
        return data

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close open dataset"""
        del self.pathfile, self.data, self.head, self._ds

    def __repr__(self):
        """print"""
        return self.head.__repr__()



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/1/16 12:06
# @Last Modified by: wqshen

import os.path
import numpy as np
import pandas as pd
from io import StringIO
from typing import Optional, Union, Any, cast
from dataclasses import dataclass, fields, astuple


@dataclass
class Diamond31Head:
    diamond: str
    dtype: int
    description: str
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond31:
    """Micaps Diamond 31 Data class

    From this class, one can read and write Micaps Diamond 31 file

    Read Diamond 31
    ---------------
    >>> m1 = Diamond31(pathfile='../../sample_data/diamond_31/18091608.000')
    >>> m1.read()
    >>> print(m1)
    >>> print(m1.data.shape)

    Write Diamond 31
    ----------------
    >>> m1 = Diamond31()
    >>> m1.head = Diamond31Head('diamond', 31, '18年09月16日08时地面填图', 2018, 9, 16, 8, 2)
    >>> m1.data = np.array([
    ...    [67297, 34.90, -19.79, 16, 32, 4, 60, 2, 142, 9999, 0, 0, 0, 32, 4, 300, 22.5, 20.0, 0, 24.2, 20, 10, 0, 0],
    ...    [96315, 114.93, 4.93, 29, 32, 7, 230, 1, 88, 9999, 2, 2, 0, 38, 1, 300, 24.8, 8.0, 5, 26.5, 24, 18]
    ...             ])
    >>> m1.write("18091608.000")
    """
    head: Optional[Diamond31Head] = None
    data: Optional[pd.DataFrame] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: Any = None, center: Any = None, stid: Any = None,
            lon: Any = None, lat: Any = None, height: Any = None, query: str = None):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        center: Any
            select data at given report center or center list (C_CCCC)
        stid: Any
            select data at given flight or flight list (C01006)
        lat: Any
            select data within given flight latitude range (V05001)
        lon: Any
            select data within given flight longitude range (V06001)
        height: Any
            select data within given flight height range (V07002)
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[[*self.index_cols, element]]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                df = df[[*self.index_cols, *element]]

        for slice_, ele_ in zip((center, stid, lat, lon, height),
                                ('C_CCCC', 'C01006', 'V05001', 'V06001', 'V07002')):
            if slice_ is not None:
                if isinstance(slice_, slice):
                    df = df[(df[ele_] >= slice_.start) & (df[ele_] <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[df[ele_].isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[df[ele_] == stid]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 5 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()

        ncols = len(self.dtype)
        headsize = len(fields(Diamond31Head))
        self.head = Diamond31Head(*data[:headsize])
        records = [','.join(data[headsize + i * ncols:headsize + (i + 1) * ncols])
                   for i in range(self.head.nrec)]
        self._ds = self.data = pd.read_csv(StringIO('\n'.join(records)))

    def write(self, pathfile: str, encoding: str = 'GBK',
              fmt: str = '%4s%5i%3i%3i%3i%7s%5i%3i%3i%5s%9.4f%9.4f%4i%4i%9.2f%8.1f%8.1f%8.2f%8.1f%8.1f%5i%5i%5i%5i%5i%5i'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str]
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head) + "\n")

            np.savetxt(f, self.data.values, fmt=fmt)

    @property
    def dtype(self):
        """column data type"""
        dtypes = [('C_CCCC', 'S4'), ('C_LY', 'i2'), ('C_LM', 'i2'), ('C_LD', 'i2'), ('C_LH', 'i2'),
                  ('C01006', 'S6'), ('V04001', 'i2'), ('V04002', 'i2'), ('V04003', 'i2'),
                  ('V_OHM', 'S4'), ('V05001', 'f4'), ('V06001', 'f4'), ('V08004', 'i2'),
                  ('V02061', 'i2'), ('V07002', 'f4'), ('V12001', 'f4'), ('V11001', 'f4'),
                  ('V11002', 'f4'), ('V11041', 'f4'), ('V11031', 'f4'), ('F07002', 'i2'),
                  ('F12001', 'i2'), ('F11001', 'i2'), ('F11002', 'i2'), ('F11041', 'i2'),
                  ('F11031', 'i2')]
        return np.dtype(dtypes)

    @property
    def names(self):
        """column-name dictionary"""
        names = [('C_CCCC', '发报中心'), ('C_LY', '年'), ('C_LM', '月'), ('C_LD', '日'),
                 ('C_LH', '时'),
                 ('C01006', '航班'), ('V04001', '观测时间年'), ('V04002', '月'),
                 ('V04003', '日'),
                 ('V_OHM', '时间'), ('V05001', '纬度'), ('V06001', '经度'),
                 ('V08004', '飞行类型'),
                 ('V02061', '导航状态'), ('V07002', '飞行高度'), ('V12001', '温度'),
                 ('V11001', '风向'),
                 ('V11001', '风速'),
                 ('V11041', '垂直速度'), ('V11031', '湍流度'),
                 ('F07002', '温度可信度'),
                 ('F12001', '风可信度'), ('F11001', '垂直速度可信度'),
                 ('F11002', '湍流度可信度'), ('F11041', '位置可信度'),
                 ('F11031', '高度可信度')]
        return dict(names)

    @property
    def index_cols(self):
        """base index columns, keep in frame when select data"""
        return ['C_CCCC', 'C_LY', 'C_LM', 'C_LD', 'C_LH', 'C01006', 'V04001', 'V04002', 'V04003',
                'V_OHM', 'V05001', 'V06001', 'V08004', 'V02061', 'V07002']

    def __repr__(self):
        """print"""
        return self.head.__repr__()

    def __getitem__(self, stid):
        """get data of station"""
        if isinstance(stid, str):
            return self.data[stid]
        elif isinstance(stid, (tuple, list, np.ndarray)):
            return self.data[np.isin(self.data['stid'], stid)]
        else:
            TypeError("stid must be type of str, tuple or list")



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/1/16 12:06
# @Last Modified by: wqshen


import os
import numpy as np
import pandas as pd
from typing import Optional, Union, Any, cast
from dataclasses import dataclass, fields, astuple
from numpy.lib.recfunctions import unstructured_to_structured


@dataclass
class Diamond42Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    minute: int
    nele: int
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond42:
    """Micaps Diamond 1 Data class

    From this class, one can read and write Micaps Diamond4 file

    Read Diamond42
    --------------
    >>> m42 = Diamond42(pathfile='../../sample_data/diamond_42/GPSG_200809162300.000')
    >>> m42.read()
    >>> print(m42)
    >>> print(m42.data.shape)
    >>> print(m42.to_frame())

    Write Diamond42
    --------------
    >>> m42 = Diamond42()
    >>> m42.head = Diamond42Head('diamond', 42, '08年07月29日20时GPS/MET数据', 2008, 7, 29, 20, 0, 8, 3)
    >>> m42.data = np.array([['ahhs', 58531, 29.714, 118.284, 142.3, 2.6846, 978.8, 23.6, 97.0,   72.7, 999999],
    ...                      ['ahma', 58336, 31.701, 118.516,  22.8, 2.6943, 993.6, 26.3, 92.0,   69.4, 999999],
    ...                      ['bais', 59211, 23.903, 106.606, 159.3, 999999, 980.6, 25.1, 84.0, 999999, 999999]
    ...             ])
    >>> m42.write("GPSG_200809162300.000")
    """
    head: Optional[Diamond42Head] = None
    data: Optional[np.ndarray] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None

    # _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: Any = None, code: Any = None, stid: Any = None, lon: Any = None,
            lat: Any = None, height: Any = None, query: str = None):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        code: Any
            select data within given code range
        stid: Any
            select data at given station or station list
        lon: Any
            select data within given longitude range
        lat: Any
            select data within given latitude range
        height: Any
            select data within given height range
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[[*self.index_cols, element]]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                df = df[[*self.index_cols, *element]]

        for slice_, ele_ in zip((code, stid, lon, lat, height), self.index_cols):
            if slice_ is not None:
                if isinstance(slice_, slice):
                    df = df[(df[ele_] >= slice_.start) & (df[ele_] <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[df[ele_].isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[df[ele_] == slice_]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 5 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()

        headsize = len(fields(Diamond42Head))
        self.head = Diamond42Head(*data[:headsize])
        ncols = len(data[headsize:]) / self.head.nrec
        if not ncols.is_integer():
            raise Exception("Data can't be splitted into integer columns.")
        self.data = np.asarray(data[headsize:]).reshape((self.head.nrec, int(ncols)))
        self._ds = self.to_frame()
    #
    def write(self, pathfile: str, encoding: str = 'GBK',
              fmt='%4s %5s %7.3f %7.3f %7.1f %6.4f %6.1f %6.1f %6.1f %6.1f %6.1f'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str], default: '%.2f'
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head[:3]) + "\n")
            f.write("\t".join(head[3:]) + "\n")

            data = self.to_numpy()
            np.savetxt(f, data, fmt=fmt)

    @property
    def dtype(self):
        dtypes = [('code', 'U4'), ('stid', 'U5'), ('lon', 'f4'), ('lat', 'f4'), ('height', 'f4'),
                  ('ztd', 'f4'), ('p', 'f4'), ('t', 'f4'), ('rh', 'f4'), ('pwv', 'f4'),
                  ('elec', 'f4')]
        return np.dtype(dtypes)

    @property
    def index_cols(self):
        """base index columns, keep in frame when select data"""
        return ['code', 'stid', 'lon', 'lat', 'height']

    def to_numpy(self):
        """Convert instance into numpy array with dtypes

        Returns
        -------
        data: np.ndarray
            converted data
        """
        return unstructured_to_structured(self.data, dtype=self.dtype)

    def to_frame(self):
        """Convert instance into pandas.DataFrame

        Returns
        -------
        data: xr.DataArray
            converted data
        """
        import pandas as pd

        data = pd.DataFrame.from_records(self.to_numpy())
        return data

    def __repr__(self):
        """print"""
        return self.head.__repr__()


if __name__ == "__main__":
    m42 = Diamond42()
    m42.head = Diamond42Head('diamond', 42, '08年07月29日20时GPS/MET数据', 2008, 7, 29, 20, 0, 8, 3)
    m42.data = np.array(
        [['ahhs', 58531, 29.714, 118.284, 142.3, 2.6846, 978.8, 23.6, 97.0, 72.7, 999999],
         ['ahma', 58336, 31.701, 118.516, 22.8, 2.6943, 993.6, 26.3, 92.0, 69.4, 999999],
         ['bais', 59211, 23.903, 106.606, 159.3, 999999, 980.6, 25.1, 84.0, 999999, 999999]
         ])
    m42.write("GPSG_200809162300.000")



#!/usr/bin/env python

# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/12/27 14:42
# @Last Modified by: wqshen

import sys
import argparse
import logzero
from logzero import logger
from pymdfs import MdfsClient


def _main():
    example_text = """Example: client_query ECMWF_HR"""

    parser = argparse.ArgumentParser(description='MDFS Data Query',
                                     epilog=example_text,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('datasource', help='data source name')
    parser.add_argument('-s', '--server', type=str, help='GDS server address',
                        default='xxx.xxx.xxx.xxx')
    parser.add_argument('-o', '--loglevel', type=int, help='loglevel: 10, 20, 30, 40, 50',
                        default=20)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    logzero.loglevel(args.loglevel)
    logger.debug(args)

    mc = MdfsClient(args.server)
    print(mc.list_vars(args.datasource))


if __name__ == '__main__':
    _main()



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/1/16 12:06
# @Last Modified by: wqshen


import re
import os
import numpy as np
from os import PathLike
from typing import Optional, Union, cast
from dataclasses import dataclass, fields


@dataclass
class Diamond14Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    fh: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond14:
    """Micaps Diamond 1 Data class

    From this class, one can read and write Micaps Diamond4 file

    Read Diamond14
    --------------
    >>> m14 = Diamond14(pathfile='../../sample_data/diamond_14/18091608.000')
    >>> m14.read()
    >>> print(m14)
    """
    head: Optional[Diamond14Head] = None
    data: Optional[dict] = None
    pathfile: Optional[str] = None
    encoding: Optional[str] = 'GBK'
    missing_value: Optional[float] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: str = None):
        """interface to select data

        Parameters
        ----------
        element: str
            select artist in data

        Returns
        -------
        data (dict): selected artist data
        """
        if element is not None and isinstance(element, str):
            return self.data[element]
        else:
            raise NotImplementedError("Only support select specific artist.")

    def read(self, path_or_string: Union[str, bytes]):
        """Read micaps diamond 14 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        """
        head, sections = self.separate_artist(path_or_string)
        self.head = Diamond14Head(*head.split())
        self.data = {
            'LINES': self.read_lines(sections['LINES']),
            'LINES_SYMBOL': self.read_lines_symbol(sections['LINES_SYMBOL']),
            'SYMBOLS': self.read_symbols(sections['SYMBOLS']),
            'CLOSED_CONTOURS': self.read_closed_contours(sections['CLOSED_CONTOURS']),
            'STATION_SITUATION': self.read_station_situation(sections['STATION_SITUATION']),
            'WEATHER_REGION': self.read_weather_region(sections['WEATHER_REGION']),
            'FILLAREA': self.read_fillarea(sections['FILLAREA']),
            'NOTES_SYMBOL': self.read_notes_symbol(sections['NOTES_SYMBOL']),
            'WithProp_LINESYMBOLS:': self.read_wthprop_linesymbols(
                sections['WithProp_LINESYMBOLS']),
        }

    def separate_artist(self, path_or_string: Union[str, bytes]):
        """Read micaps diamond 14 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        """
        if isinstance(path_or_string, bytes):
            s = path_or_string.decode(self.encoding)
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=self.encoding) as f:
                s = f.read()
        else:
            s = path_or_string

        segments = re.split(
            r'(\nLINES:|\nLINES_SYMBOL:|\nSYMBOLS:|\nCLOSED_CONTOURS:|\nSTATION_SITUATION|'
            r'\nWEATHER_REGION:|\nFILLAREA:|\nNOTES_SYMBOL:|\nWithProp_LINESYMBOLS:)',
            s)
        head = segments[0]

        n = len(segments) - 1
        if n % 2 != 0:
            raise Exception("re.split failed to separate key and values")
        sections = {
            segments[i * 2 + 1].split(':')[0][1:]: segments[i * 2 + 1][1:] + segments[i * 2 + 2]
            for i in range(int(n / 2))
        }
        return head, sections

    def read_lines(self, content: str):
        data = content.split()
        nlines = int(data[1])
        if nlines > 0:
            # define data
            line_width = []
            line_xyz_num = []
            line_xyz = []
            line_label_num = []
            line_label = []
            line_label_xyz = []
            idx = 2
            for i in range(nlines):
                # line width
                width = float(data[idx])
                line_width.append(width)
                idx += 1

                # line xyz point number
                xyz_num = int(data[idx])
                line_xyz_num.append(xyz_num)
                idx += 1

                # line xyz
                xyz = np.array(data[idx:(idx + 3 * xyz_num)]).astype(np.float)
                xyz.shape = [xyz_num, 3]
                line_xyz.append(xyz)
                idx += xyz_num * 3

                # line label
                label = data[idx]
                line_label.append(label)
                idx += 1

                # line label number
                label_num = int(data[idx])
                line_label_num.append(label_num)
                idx += 1

                # label xyz
                if label_num > 0:
                    label_xyz = np.array(
                        data[idx:(idx + 3 * label_num)]).astype(np.float)
                    label_xyz.shape = [label_num, 3]
                    line_label_xyz.append(label_xyz)
                    idx += label_num * 3
                else:
                    line_label_xyz.append([])

            # construct line data type
            lines = {
                "line_width": line_width, "line_xyz_num": line_xyz_num,
                "line_xyz": line_xyz, "line_label_num": line_label_num,
                "line_label": line_label, "line_label_xyz": line_label_xyz}
        else:
            lines = {}
        return lines

    def read_lines_symbol(self, content):
        data = content.split()
        nlines = int(data[1])
        idx = 2

        # loop every line symbol
        if nlines > 0:
            # define data
            linesym_code = []
            linesym_width = []
            linesym_xyz_num = []
            linesym_xyz = []

            for i in range(nlines):
                # line symbol code
                code = int(data[idx])
                linesym_code.append(code)
                idx += 1

                # line width
                width = float(data[idx])
                linesym_width.append(width)
                idx += 1

                # line symbol xyz point number
                xyz_num = int(data[idx])
                linesym_xyz_num.append(xyz_num)
                idx += 1

                # line symbol xyz
                xyz = np.array(data[idx:(idx + 3 * xyz_num)]).astype(np.float)
                xyz.shape = [xyz_num, 3]
                linesym_xyz.append(xyz)
                idx += xyz_num * 3

                # line symbol label
                label = data[idx]
                idx += 1

                # line symbol label number
                label_num = int(data[idx])
                idx += label_num * 3 + 1

            lines_symbol = {"linesym_code": linesym_code,
                            "linesym_width": linesym_width,
                            "linesym_xyz_num": linesym_xyz_num,
                            "linesym_xyz": linesym_xyz}
        else:
            lines_symbol = {}
        return lines_symbol

    def read_symbols(self, content):
        data = content.split()
        nlines = int(data[1])
        idx = 2

        # loop every symbol
        if nlines > 0:
            # define data
            symbol_code = []
            symbol_xyz = []
            symbol_value = []

            for i in range(nlines):
                # symbol code
                code = int(data[idx])
                symbol_code.append(code)
                idx += 1

                # symbol xyz
                xyz = np.array(data[idx:(idx + 3)]).astype(np.float)
                symbol_xyz.append(xyz)
                idx += 3

                # symbol value
                value = data[idx]
                symbol_value.append(value)
                idx += 1

            symbols = {"symbol_code": symbol_code,
                       "symbol_xyz": symbol_xyz,
                       "symbol_value": symbol_value}
        else:
            symbols = {}
        return symbols

    def read_closed_contours(self, content):
        data = content.split()
        nlines = int(data[1])
        idx = 2
        # loop every closed contour
        if nlines > 0:
            # define data
            cn_width = []
            cn_xyz_num = []
            cn_xyz = []
            cn_label_num = []
            cn_label = []
            cn_label_xyz = []

            for i in range(nlines):
                # line width
                width = float(data[idx])
                cn_width.append(width)
                idx += 1

                # line xyz point number
                xyz_num = int(data[idx])
                cn_xyz_num.append(xyz_num)
                idx += 1

                # line xyz
                xyz = np.array(data[idx:(idx + 3 * xyz_num)]).astype(np.float)
                xyz.shape = [xyz_num, 3]
                cn_xyz.append(xyz)
                idx += 3 * xyz_num

                # line label
                label = data[idx]
                cn_label.append(label)
                idx += 1

                # line label number
                label_num = int(data[idx])
                cn_label_num.append(label_num)
                idx += 1

                # label xyz
                if label_num > 0:
                    label_xyz = np.array(
                        data[idx:(idx + 3 * label_num)]).astype(np.float)
                    label_xyz.shape = [3, label_num]
                    cn_label_xyz.append(label_xyz)
                    idx += label_num * 3
                else:
                    cn_label_xyz.append([])

            closed_contours = {
                "cn_width": cn_width, "cn_xyz_num": cn_xyz_num,
                "cn_xyz": cn_xyz, "cn_label": cn_label,
                "cn_label_num": cn_label_num, "cn_label_xyz": cn_label_xyz}
        else:
            closed_contours = {}
        return closed_contours

    def read_station_situation(self, content):
        data = content.split()
        if len(data) > 1:
            stations = np.array(data[1:])
            stations.shape = [len(stations) // 2, 2]
        else:
            stations = None
        return stations

    def read_weather_region(self, content):
        data = content.split()
        nlines = int(data[1])
        idx = 2

        # loop every region
        if nlines > 0:
            # define data
            weather_region_code = []
            weather_region_xyz_num = []
            weather_region_xyz = []

            for i in range(nlines):
                # region code
                code = int(data[idx])
                weather_region_code.append(code)
                idx += 1

                # region xyz point number
                xyz_num = int(data[idx])
                weather_region_xyz_num.append(xyz_num)
                idx += 1

                # region xyz point
                xyz = np.array(
                    data[idx:(idx + 3 * xyz_num)]).astype(np.float)
                xyz.shape = [xyz_num, 3]
                weather_region_xyz.append(xyz)
                idx += 3 * xyz_num

            weather_region = {
                "weather_region_code": weather_region_code,
                "weather_region_xyz_num": weather_region_xyz_num,
                "weather_region_xyz": weather_region_xyz}
        else:
            weather_region = {}
        return weather_region

    def read_fillarea(self, content):
        data = content.split()
        nlines = int(data[1])
        idx = 2

        # loop every fill area
        if nlines > 0:
            # define data
            fillarea_code = []
            fillarea_num = []
            fillarea_xyz = []
            fillarea_type = []
            fillarea_color = []
            fillarea_frontcolor = []
            fillarea_backcolor = []
            fillarea_gradient_angle = []
            fillarea_graphics_type = []
            fillarea_frame = []

            for i in range(nlines):
                # code
                code = int(data[idx])
                fillarea_code.append(code)
                idx += 1

                # xyz point number
                xyz_num = int(data[idx])
                fillarea_num.append(xyz_num)
                idx += 1

                # xyz point
                xyz = np.array(
                    data[idx:(idx + 3 * xyz_num)]).astype(np.float)
                xyz.shape = [xyz_num, 3]
                fillarea_xyz.append(xyz)
                idx += 3 * xyz_num

                # fill type
                ftype = int(data[idx])
                fillarea_type.append(ftype)
                idx += 1

                # line color
                color = np.array(data[idx:(idx + 4)]).astype(np.int)
                fillarea_color.append(color)
                idx += 4

                # front color
                front_color = np.array(data[idx:(idx + 4)]).astype(np.int)
                fillarea_frontcolor.append(front_color)
                idx += 4

                # background color
                back_color = np.array(data[idx:(idx + 4)]).astype(np.int)
                fillarea_backcolor.append(back_color)
                idx += 4

                # color gradient angle
                gradient_angle = float(data[idx])
                fillarea_gradient_angle.append(gradient_angle)
                idx += 1

                # graphics type
                graphics_type = int(data[idx])
                fillarea_graphics_type.append(graphics_type)
                idx += 1

                # draw frame or not
                frame = int(data[idx])
                fillarea_frame.append(frame)
                idx += 1

            fill_area = {
                "fillarea_code": fillarea_code, "fillarea_num": fillarea_num,
                "fillarea_xyz": fillarea_xyz, "fillarea_type": fillarea_type,
                "fillarea_color": fillarea_color,
                "fillarea_frontcolor": fillarea_frontcolor,
                "fillarea_backcolor": fillarea_backcolor,
                "fillarea_gradient_angle": fillarea_gradient_angle,
                "fillarea_graphics_type": fillarea_graphics_type,
                "fillarea_frame": fillarea_frame}
        else:
            fill_area = {}
        return fill_area

    def read_notes_symbol(self, content):
        data = content.split()
        nlines = int(data[1])
        idx = 2

        # loop every notes symbol
        if nlines > 0:
            # define data
            nsymbol_code = []
            nsymbol_xyz = []
            nsymbol_charLen = []
            nsymbol_char = []
            nsymbol_angle = []
            nsymbol_fontLen = []
            nsymbol_fontName = []
            nsymbol_fontSize = []
            nsymbol_fontType = []
            nsymbol_color = []

            for i in range(nlines):
                # code
                code = int(data[idx])
                nsymbol_code.append(code)
                idx += 1

                # xyz
                xyz = np.array(data[idx:(idx + 3)]).astype(np.float)
                nsymbol_xyz.append([xyz])
                idx += 3

                # character length
                char_len = int(data[idx])
                nsymbol_charLen.append(char_len)
                idx += 1

                # characters
                char = data[idx]
                nsymbol_char.append(char)
                idx += 1

                # character angle
                angle = data[idx]
                nsymbol_angle.append(angle)
                idx += 1

                # font length
                font_len = data[idx]
                nsymbol_fontLen.append(font_len)
                idx += 1

                # font name
                font_name = data[idx]
                nsymbol_fontName.append(font_name)
                idx += 1

                # font size
                font_size = data[idx]
                nsymbol_fontSize.append(font_size)
                idx += 1

                # font type
                font_type = data[idx]
                nsymbol_fontType.append(font_type)
                idx += 1

                # color
                color = np.array(data[idx:(idx + 4)]).astype(np.int)
                nsymbol_color.append(color)
                idx += 4

            notes_symbol = {
                "nsymbol_code": nsymbol_code,
                "nsymbol_xyz": nsymbol_xyz,
                "nsymbol_charLen": nsymbol_charLen,
                "nsymbol_char": nsymbol_char,
                "nsymbol_angle": nsymbol_angle,
                "nsymbol_fontLen": nsymbol_fontLen,
                "nsymbol_fontName": nsymbol_fontName,
                "nsymbol_fontSize": nsymbol_fontSize,
                "nsymbol_fontType": nsymbol_fontType,
                "nsymbol_color": nsymbol_color}
        else:
            notes_symbol = {}
        return notes_symbol

    def read_wthprop_linesymbols(self, content):
        data = content.split()
        nlines = int(data[1])
        idx = 2

        # loop every line symbol
        if nlines > 0:
            # define data
            plinesym_code = []
            plinesym_width = []
            plinesym_color = []
            plinesym_type = []
            plinesym_shadow = []
            plinesym_xyz_num = []
            plinesym_xyz = []
            plinesym_label = []
            plinesym_label_num = []
            plinesym_label_xyz = []

            for i in range(nlines):
                # line symbol code
                code = int(data[idx])
                plinesym_code.append(code)
                idx += 1

                # line width
                width = float(data[idx])
                plinesym_width.append(width)
                idx += 1

                # line color
                color = np.array(data[idx:(idx + 4)]).astype(np.int)
                plinesym_color.append([color])
                idx += 4

                # line type
                ltype = int(data[idx])
                plinesym_type.append(ltype)
                idx += 1

                # line shadow
                shadow = int(data[idx])
                plinesym_shadow.append(shadow)
                idx += 1

                # line symbol xyz point nlines
                xyz_num = int(data[idx])
                plinesym_xyz_num.append(xyz_num)
                idx += 1

                # line symbol xyz
                xyz = np.array(data[idx:(idx + 3 * xyz_num)]).astype(np.float)
                xyz.shape = [xyz_num, 3]
                plinesym_xyz.append(xyz)
                idx += 3 * xyz_num

                # line symbol label
                label = data[idx]
                plinesym_label.append(label)
                idx += 1

                # line label nlines
                label_num = int(data[idx])
                plinesym_label_num.append(label_num)
                idx += 1

                # label xyz
                if label_num > 0:
                    label_xyz = np.array(
                        data[idx:(idx + 3 * label_num)]).astype(np.float)
                    label_xyz.shape = [label_num, 3]
                    plinesym_label_xyz.append(label_xyz)
                    idx += label_num * 3
                else:
                    plinesym_label_xyz.append([])

            plines_symbol = {
                "plinesym_code": plinesym_code,
                "plinesym_width": plinesym_width,
                "plinesym_color": plinesym_color,
                "plinesym_type": plinesym_type,
                "plinesym_shadow": plinesym_shadow,
                "plinesym_xyz_num": plinesym_xyz_num,
                "plinesym_xyz": plinesym_xyz,
                "plinesym_label": plinesym_label,
                "plinesym_label_num": plinesym_label_num,
                "plinesym_label_xyz": plinesym_label_xyz}
        else:
            plines_symbol = {}
        return plines_symbol

    def __repr__(self):
        """print"""
        return self.head.__repr__()

    def __getitem__(self, artist: str):
        """get data of station"""
        return self.data[artist]



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/1/16 12:06
# @Last Modified by: wqshen

import os
import numpy as np
import pandas as pd
from typing import Optional, Union, Any, cast
from dataclasses import dataclass, fields, astuple
from numpy.lib.recfunctions import unstructured_to_structured


@dataclass
class Diamond41Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond41:
    """Micaps Diamond 41 Data class

    From this class, one can read and write Micaps Diamond 41 file

    Read Diamond 41
    ---------------
    >>> m41 = Diamond41(pathfile='../../sample_data/diamond_41/LIGHT_MICAPS3_2008062417_COLLECT.TXT')
    >>> m41.read()
    >>> print(m41)
    >>> print(m41.data.shape)
    >>> print(m41.to_frame())

    Write Diamond 41
    ----------------
    >>> m41 = Diamond41()
    >>> m41.head = Diamond41Head('diamond', 41, '2008年06月24日17时闪电监测资料', 2008, 1, 8, 8)
    >>> m41.data = np.array([
    ...    [1, 200806241711032294853, 9999, 9999, 114.8664, 26.57132, 9999, 9999, 9999, 9999, -29.16124, 9999, 74.63686, 6, -7.178152],
    ...    [2, 200806241711036436922, 9999, 9999, 106.4083, 23.33219, 9999, 9999, 9999, 9999, -51.30987, 9999, 75.52758, 6, -9.658328],
    ...    [3, 200806241711045918992, 9999, 9999, 121.0393, 28.38015, 9999, 9999, 9999, 9999, -28.92107, 9999, 74.09919, 6, -6.855365],
    ...    [4, 200806241711049273112, 9999, 9999, 102.8865, 25.18426, 9999, 9999, 9999, 9999, -205.6933, 9999, 0, 5, -33.32752],
    ...    [5, 20080624171105456599, 9999, 9999, 110.0666, 24.17627, 9999, 9999, 9999, 9999, -46.67589, 9999, 0, 2, -13.2767],
    ...    [6, 200806241711053232880, 9999, 9999, 109.5917, 24.25497, 9999, 9999, 9999, 9999, -31.43308, 9999, 0, 2, -10.58799],
    ...             ])
    >>> m41.write("LIGHT_MICAPS3_2008010808_COLLECT.TXT")
    """
    head: Optional[Diamond41Head] = None
    data: Optional[np.ndarray] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: Any = None, lon: Any = None, lat: Any = None, height: Any = None,
            query: str = None):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        lon: Any
            select data within given longitude range
        lat: Any
            select data within given latitude range
        height: Any
            select data within given height range
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[[*self.index_cols, element]]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                df = df[[*self.index_cols, *element]]

        for slice_, ele_ in zip((lon, lat, height), ('lon', 'lat', 'height')):
            if slice_ is not None:
                if isinstance(slice_, slice):
                    df = df[(df[ele_] >= slice_.start) & (df[ele_] <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[df[ele_].isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[df[ele_] == slice_]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 5 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()

        headsize = len(fields(Diamond41Head))
        self.head = Diamond41Head(*data[:headsize])

        ncols = len(data[headsize:]) / self.head.nrec
        if not ncols.is_integer():
            raise Exception("Data can't be splitted into integer columns.")
        self.data = np.asarray(data[headsize:]).reshape((self.head.nrec, int(ncols)))
        self._ds = self.to_frame()

    def write(self, pathfile: str, encoding: str = 'GBK',
              fmt: str = '%5i %22s %5i %5i %9.4f %9.4f %9.4f %5i %9.4f %9.4f %9.5f %9.4f %9.5f %5i %9.5f'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str]
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head[:3]) + "\n")
            f.write("\t".join(head[3:8]) + "\n")

            data = self.to_numpy()
            np.savetxt(f, data, fmt=fmt)

    @property
    def dtype(self):
        """column data type"""
        dtypes = [('sn', 'i4'), ('datetime', 'U21'), ('unit_code', 'i4'), ('category', 'i1'),
                  ('lon', 'f8'), ('lat', 'f8'), ('height', 'f4'),
                  ('return_stroke', 'i4'), ('rise_time', 'f4'), ('decay_time', 'f4'),
                  ('intensity', 'f4'), ('energy', 'f4'), ('error', 'f4'),
                  ('algorithm', 'i4'), ('steepness', 'f4')]
        return np.dtype(dtypes)

    @property
    def names(self):
        """column-name dictionary"""
        names = [('sn', '闪电个数的序号'), ('datetime', '日期时间'), ('unit_code', '单位代码'),
                 ('category', '闪电的种类'), ('lon', '闪电位置的经度'), ('lat', '闪电位置的纬度'),
                 ('height', '闪电位置的高度'),
                 ('return_stroke', '闪电回击数'), ('rise_time', '上升时间'),
                 ('decay_time', '衰减时间'),
                 ('intensity', '闪电归一化电流强度值'), ('energy', '闪电能量'), ('error', '误差'),
                 ('algorithm', '算法代码'), ('steepness', '陡度')]
        return dict(names)

    @property
    def index_cols(self):
        """base index columns, keep in frame when select data"""
        return ['sn', 'datetime', 'unit_code', 'category', 'lon', 'lat', 'height']

    def to_numpy(self):
        """Convert instance into numpy array with dtypes

        Returns
        -------
        data: np.ndarray
            converted data
        """
        return unstructured_to_structured(self.data, dtype=self.dtype)

    def to_frame(self):
        """Convert instance into pandas.DataFrame

        Returns
        -------
        data: xr.DataArray
            converted data
        """
        data = pd.DataFrame.from_records(self.to_numpy())
        return data

    def __repr__(self):
        """print"""
        return self.head.__repr__()

    def __getitem__(self, sn):
        """get data of station"""
        if isinstance(sn, str):
            return self.data[sn]
        elif isinstance(sn, (tuple, list, np.ndarray)):
            return self.data[np.isin(self.data['sn'], sn)]
        else:
            TypeError("stid must be type of str, tuple or list")



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2023/2/9 22:07
# @Last Modified by: wqshen


import os
import yaml
import itertools
import numpy as np
import pandas as pd
from typing import Optional, Union, Any, cast
from struct import unpack, calcsize
from collections import OrderedDict
from dataclasses import dataclass, fields


@dataclass
class MdfsStationHead:
    discriminator: str
    dtype: int
    description: str
    level: float
    levelDescription: str
    year: int
    month: int
    day: int
    hour: int
    minute: int
    second: int
    timezone: int
    Extent: str

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            if isinstance(field_value, bytes):
                field_value = field_value.decode('GBK').rstrip('\x00')
            setattr(self, field.name, field.type(field_value))


@dataclass
class MdfsStationData(object):
    head: Optional[MdfsStationHead] = None
    stationNum: Optional[int] = None
    elementNum: Optional[int] = None
    elementIdTypeMap: Optional[OrderedDict] = None
    data: Optional[dict] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)
            self._ds = self.to_dataframe()

    def sel(self, element: Any = None, query: str = None, **kwargs):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        kwargs: dict
            select data key within given level range
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[self.translate(element)]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                element = [self.translate(ele) for ele in element]
                df = df[[*element]]

        for ele_, slice_ in kwargs.items():
            ele_ = self.translate(ele_)
            if slice_ is not None:
                if isinstance(slice_, slice):
                    df = df[(df[ele_] >= slice_.start) & (df[ele_] <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[df[ele_].isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[df[ele_] == slice_]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def translate(self, ele):
        """translate element name into column name"""
        key_map = {'stid': '站号', 'lon': '经度', 'lat': '纬度', 'height': '测站高度',
                   'level': '测站级别（short）'}
        return key_map.get(ele, ele)

    def read(self, path_or_bytes: Union[str, bytes]):
        if isinstance(path_or_bytes, str):
            with open(path_or_bytes, 'rb') as f:
                bytes_array = f.read()
        else:
            bytes_array = path_or_bytes

        head_data = unpack('=4sh100sf50siiiiiii100s', bytes_array[:288])
        self.head = MdfsStationHead(*head_data)

        stationNum, elementNum = unpack("=ih", bytes_array[288:288 + 6])
        elementIdTypeMap = unpack("={}h".format(elementNum * 2),
                                  bytes_array[294:294 + 4 * elementNum])
        elementIdTypeMap = OrderedDict(
            itertools.zip_longest(*[iter(elementIdTypeMap)] * 2, fillvalue="")
        )
        self.stationNum = stationNum
        self.elementNum = elementNum
        self.elementIdTypeMap = elementIdTypeMap

        p = 294 + 4 * elementNum
        data = {}
        _id_dict = self._load_id_description()
        for i in range(stationNum):
            data_head = list(unpack("=iffh", bytes_array[p:p + 14]))
            p += 14

            element = {"站号": data_head[0], "经度": data_head[1], "纬度": data_head[2]}
            for j in range(data_head[-1]):
                element_id = unpack("=h", bytes_array[p:p + 2])[0]
                p += 2
                element_type = self._element_type_dict[elementIdTypeMap[element_id]]
                elemet_value = \
                    unpack("=" + element_type, bytes_array[p:p + calcsize("=" + element_type)])[0]
                element[_id_dict.get(element_id, element_id)] = elemet_value
                p += calcsize("=" + element_type)
                data[data_head[0]] = element
        self.data = data
        self._ds = self.to_dataframe()

    @property
    def _element_type_dict(self) -> dict:
        return {1: 'b', 2: 'h', 3: 'i', 4: 'l', 5: 'f', 6: 'd', 7: 's'}

    def _load_id_description(self, pathfile: str = '../config/ElementNameDict.yml'):
        if not os.path.isfile(pathfile):
            path = os.path.dirname(os.path.realpath(__file__))
            pathfile = os.path.join(path, pathfile)
            if not os.path.isfile(pathfile):
                raise FileNotFoundError(f"configuration file {pathfile} not found")

        with open(pathfile, 'r', encoding='utf8') as f:
            return yaml.load(f, Loader=yaml.SafeLoader)

    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame.from_dict(self.data, orient='index')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close open dataset"""
        del self.pathfile, self.data, self.head, self._ds

    def __repr__(self):
        """print"""
        return self.head.__repr__()



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2021/8/4 20:20
# @Last Modified by: wqshen

import os
import numpy as np
import pandas as pd
from typing import Optional, Union, Any, cast
from dataclasses import dataclass, fields, astuple
from numpy.lib.recfunctions import unstructured_to_structured


@dataclass
class Diamond7Head:
    diamond: str
    dtype: int
    description: str

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond7Index:
    name: str
    sn: str
    center: str
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond7:
    """Micaps Diamond 7 Data class

    From this class, one can read and write Micaps Diamond7 file

    Read Diamond5
    --------------
    >>> m7 = Diamond7(pathfile="./micaps/typhoon/ECEP/ECEP14W0408.dat")
    >>> m7.read()
    >>> print(m7)
    >>> print(m7.to_frame())

    Write Diamond5
    --------------
    >>> m5 = Diamond7()
    >>> m5.head = Diamond7Head('diamond', 7, '9714号台风路径(主观预报)', 'NAMELESS', '0615', '28', 3)
    >>> m5.data = np.array([
    ...    [2006, 09, 24, 08, 00, 111.3, 15.9, 995, 18, 100.0, 0.0, 292.5, 15.0],
    ...    [2006, 09, 24, 08, 24, 107.4, 17.3, 985, 23, 0.0, 0.0, 0.0, 0.0]
    ...             ])
    >>> m5.write("18091608.000")
    """
    head: Optional[Diamond7Head] = None
    index: Optional[pd.DataFrame] = None
    data: Optional[list] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = 9999
    _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: Any = None, name: Any = None, sn: Any = None, center: Any = None,
            query: str = None):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        name: Any
            select data at given station or station list
        sn: Any
            select data within given longitude range
        center: Any
            select data within given latitude range
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[[*self.index_cols, element]]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                df = df[[*self.index_cols, *element]]

        for slice_, ele_ in zip((name, sn, center), self.index_cols):
            if slice_ is not None:
                idx = df.index.get_level_values(ele_)
                if isinstance(slice_, slice):
                    df = df[(idx >= slice_.start) & (idx <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[idx.isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[idx == slice_]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 7 file

        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file

        Notes
        -----
        Before call this method, ensure you have set the `pathfile`
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split('\n')
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.readlines()
        else:
            data = path_or_string.split('\n')

        indices, tracks, track = [], [], []
        headsize = len(fields(Diamond7Head))
        head = data[0].split()
        if len(head) != headsize:
            raise ValueError("Field in first line has inconsistent size to HEAD.")
        else:
            self.head = Diamond7Head(*head)

        for row in data[1:]:
            row_data = row.split()
            if row_data == ['0']:
                tracks.append(track)
                track = []
            else:
                track.append(tuple(row_data))
        del track
        if len(tracks) == 0:
            print("Warning: Failed to find any track in file.")
            return
        else:
            index = unstructured_to_structured(np.array([track[0] for track in tracks]),
                                               dtype=self.dtype_index)
            self.index = pd.DataFrame.from_records(index)
            self.data = [np.asarray(track[1:], dtype=self.dtype) for track in tracks]
        self._ds = self.to_frame()

    def write(self, pathfile: str, encoding: str = 'GBK',
              fmt='%4i %02i %02i %02i %03i %6.1f %5.1f %4.0f %3.0f %5.0f %5.0f %5.0f %5.0f'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str], default: '%.2f'
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head) + "\n")
            for i, (head, track) in enumerate(zip(self.index.values, self.data)):
                if i > 0:
                    f.write('0\n')
                f.write("\t".join(map(str, head)) + "\n")
                np.savetxt(f, track, fmt=fmt)

    @property
    def dtype(self):
        return np.dtype({'names': ['year', 'month', 'day', 'hour', 'fh',
                                   'lon', 'lat', 'wind', 'pres',
                                   'r7', 'r10', 'dir', 'spd'],
                         'formats': ['i2', 'i1', 'i1', 'i1', 'i2', 'f4',
                                     'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4']})

    @property
    def dtype_index(self):
        return np.dtype({'names': ['name', 'sn', 'center', 'nrec'],
                         'formats': ['U50', 'U4', 'U4', 'i1']})

    @property
    def index_cols(self):
        """base index columns, keep in frame when select data"""
        return ['name', 'sn', 'center']

    def to_frame(self):
        """Convert instance into pandas.DataFrame

        Returns
        -------
        data: xr.DataArray
            converted data
        """
        nrecs = self.index.nrec
        multi_index = self.index.loc[self.index.index.repeat(nrecs)]
        multi_index = pd.MultiIndex.from_frame(multi_index[self.index_cols])
        data = pd.DataFrame.from_records(np.hstack(self.data), index=multi_index)
        return data

    def __repr__(self):
        """print"""
        return self.head.__repr__()

    def __getitem__(self, sn):
        """get data of station"""
        if isinstance(sn, str):
            return self._ds[self._ds.sn == sn]
        elif isinstance(sn, (tuple, list, np.ndarray)):
            return self._ds[self._ds.sn.isin(sn)]
        else:
            TypeError("stid must be type of str, tuple or list")



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/1/16 12:06
# @Last Modified by: wqshen


import os
import numpy as np
import pandas as pd
from typing import Optional, Union, Any, cast
from dataclasses import dataclass, fields, astuple
from numpy.lib.recfunctions import unstructured_to_structured


@dataclass
class Diamond3Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    level: int

    # if nlines is 0, you should set lines=[]
    nlines: int
    lines: list
    smooth: float
    boldvalue: float

    # if nedgepoints is 0, you should set edgepoints=[]
    nedgepoints: int
    edgepoints: list

    nele: int
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond3:
    """Micaps Diamond 3 (通用填图和离散点等值线) Data class

    From this class, one can read and write Micaps Diamond 3 file

    Read Diamond 3
    --------------
    >>> m3 = Diamond3(pathfile='../../sample_data/diamond_3/18091608.000')
    >>> m3.read()
    >>> print(m3)
    >>> print(m3.data.shape)
    >>> print(m3.to_frame())

    Write Diamond 3
    ---------------
    >>> m3 = Diamond3()
    >>> m3.head = Diamond3Head('diamond', 3, '高空700hPa变温_20210725200000',
    ...                         2021, 7, 25, 20, 700, 0, [], 0, 0, 0, [], 1, 487)
    >>> m3.data = np.array([
    ...    [47104, 128.85, 37.8, 80, 0.6],
    ...    [83971, -51.18, -30, 3, 4.6],
    ...    [29698, 99.03, 54.88, 411, 0.0],
    ...    [76805, -99.75, 16.75, 3, 0.2],
    ...    [36872, 77, 43.35, 662, 0.4],
    ...             ])
    >>> m3.write("2021072520.000")
    """
    head: Optional[Diamond3Head] = None
    data: Optional[np.ndarray] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: Any = None, stid: Any = None, lon: Any = None, lat: Any = None,
            height: Any = None, query: str = None):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        stid: Any
            select data at given station or station list
        lon: Any
            select data within given longitude range
        lat: Any
            select data within given latitude range
        height: Any
            select data within given height range
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[[*self.index_cols, element]]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                df = df[[*self.index_cols, *element]]

        for slice_, ele_ in zip((stid, lon, lat, height), self.index_cols):
            if slice_ is not None:
                if isinstance(slice_, slice):
                    df = df[(df[ele_] >= slice_.start) & (df[ele_] <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[df[ele_].isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[df[ele_] == slice_]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 3 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()
        headsize = len(fields(Diamond3Head))
        head = data[:9]
        nlines = int(head[-1])
        head += [data[9:9 + nlines]]
        head += data[9 + nlines:12 + nlines]
        nedgepoints = int(head[-1])
        head += [data[12 + nlines:12 + nlines + nedgepoints]]
        head += data[12 + nlines + nedgepoints:14 + nlines + nedgepoints]
        self.head = Diamond3Head(*head)

        dynamic_headsize = headsize + (nlines - 1) + (nedgepoints - 1)
        ncols = len(data[dynamic_headsize:]) / self.head.nrec
        if not ncols.is_integer():
            raise Exception("Data can't be splitted into integer columns.")
        self.data = np.asfarray(data[dynamic_headsize:]).reshape((self.head.nrec, int(ncols)))
        self._ds = self.to_frame()

    def write(self, pathfile: str, encoding: str = 'GBK', fmt: str = '%5i%8.2f%8.2f%5i%8.1f'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str], default: '%5i%8.2f%8.2f%5i%8.1f'
            format to save numpy data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            lines, edgepoints = map(str, self.head.lines), map(str, self.head.edgepoints)
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head[:8]) + "\n")
            f.write(head[8] + "\t" + "\t".join(lines) + "\t" + "\t".join(head[10:12]) + "\n")
            f.write(head[12] + "\t" + "\t".join(edgepoints) + "\n")
            f.write("\t".join(head[-2:]) + "\n")

            data = self.to_numpy()
            ffields = fmt.split('%')
            if self.data.shape[-1] > len(ffields):
                fmt += ffields[-1] * (self.data.shape[-1] - len(ffields))
            np.savetxt(f, data, fmt=fmt)

    @property
    def dtype(self):
        """column data type"""
        ncols = self.data.shape[-1]
        dtypes = [('stid', 'i4'), ('lon', 'f4'), ('lat', 'f4'), ('height', 'f4'), ('ele1', 'f4')]
        if ncols > len(dtypes):
            dtypes += [(f'ele{i + 2}', str) for i in range(ncols - len(dtypes))]
        return np.dtype(dtypes)

    @property
    def names(self):
        """column-name dictionary"""
        names = [('stid', '区站号'), ('lon', '经度'), ('lat', '纬度'), ('height', '拔海高度'),
                 ('ele1', '要素值1')]
        ncols = self.data.shape[-1]
        if ncols > len(names):
            names += [(f'ele{i + 2}', f'要素值{i + 2}') for i in range(ncols - len(names))]
        return dict(names)

    @property
    def index_cols(self):
        """base index columns, keep in frame when select data"""
        return ['stid', 'lon', 'lat', 'height']

    def to_numpy(self):
        """Convert instance into numpy array with dtypes

        Returns
        -------
        data: np.ndarray
            converted data
        """
        return unstructured_to_structured(self.data, dtype=self.dtype)

    def to_frame(self):
        """Convert instance into pandas.DataFrame

        Returns
        -------
        data: pd.DataFrame
            converted data
        """
        data = pd.DataFrame.from_records(self.to_numpy())
        return data

    def __repr__(self):
        """print"""
        return self.head.__repr__()

    def sel(self, **kwargs):
        """interface to select variable from file by given more filter

        Parameters
        ----------
        kwargs (dict): other parameters used to filter data

        Returns
        -------
        var (pandas.DataFrame): Readed variable in pandas.DataFrame
        """
        if len(kwargs) > 0:
            conditions = True
            for k, v in kwargs.items():
                if k in ('stid',):
                    conditions = conditions & (self._ds[k] == v)
                elif k in ('lat', 'lon', 'height'):
                    conditions = conditions & (self._ds[k].between(v.start, v.stop))
            return self._ds.loc[conditions]
        else:
            return self._ds



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Date: 2019/4/6 11:59
# @Last Modified by: wqshen


from setuptools import setup, find_packages

version = "0.1"
name = "pymdfs"

install_requires = [
    "awx",
    "numpy>=1.10",
    "xarray",
    "pandas",
    "logzero>=1.0",
    "pyyaml",
    "protobuf",
    "retrying",
    "httplib2"
]

classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research/Operation",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Topic :: GeoScience :: Data",
]

setup(
    name=name,
    version=version,
    description="High level python interface to Micaps MDFS Data",
    author="Wenqiang Shen",
    author_email="wqshen91@163.com",
    python_requires=">=3.7",
    install_requires=install_requires,
    packages=find_packages(),
    long_description=open('README.rst', 'r', encoding='utf8').read(),
    url='https://github.com/zjobsdev/PyMDFS',
    entry_points={
        'console_scripts': [
            'mdfs_dump=pymdfs.client_dump:_main',
            'mdfs_query=pymdfs.client_query:_main',
        ]
    },
    include_package_data=True,
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2019/1/16 12:06
# @Last Modified by: wqshen

import os
import numpy as np
import pandas as pd
from typing import Optional, Union, Any, cast
from dataclasses import dataclass, fields, astuple
from numpy.lib.recfunctions import unstructured_to_structured


@dataclass
class Diamond1Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond1:
    """Micaps Diamond 1 (地面全要素填图数据) Data class

    From this class, one can read and write Micaps Diamond 1 file

    Read Diamond1
    --------------
    >>> m1 = Diamond1(pathfile='../../sample_data/diamond_1/18091608.000')
    >>> m1.read()
    >>> print(m1)
    >>> print(m1.data.shape)
    >>> print(m1.to_frame())

    Write Diamond1
    --------------
    >>> m1 = Diamond1()
    >>> m1.head = Diamond1Head('diamond', 1, '18年09月16日08时地面填图', 2018, 9, 16, 8, 2)
    >>> m1.data = np.array([
    ...    [67297, 34.90, -19.79, 16, 32, 4, 60, 2, 142, 9999,
    ...     0, 0, 0, 32, 4, 300, 22.5, 20.0, 0, 24.2, 20, 10, 0, 0],
    ...    [96315, 114.93, 4.93, 29, 32, 7, 230, 1, 88, 9999,
    ...     2, 2, 0, 38, 1, 300, 24.8, 8.0, 5, 26.5, 24, 18]
    ...             ])
    >>> m1.write("18091608.000")
    """
    head: Optional[Diamond1Head] = None
    data: Optional[np.ndarray] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: Any = None, stid: Any = None, lon: Any = None, lat: Any = None,
            height: Any = None, level: Any = None, query: str = None):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        stid: Any
            select data at given station or station list
        lon: Any
            select data within given longitude range
        lat: Any
            select data within given latitude range
        height: Any
            select data within given height range
        level: Any
            select data within given level range
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[[*self.index_cols, element]]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                df = df[[*self.index_cols, *element]]

        for slice_, ele_ in zip((stid, lon, lat, height, level), self.index_cols):
            if slice_ is not None:
                if isinstance(slice_, slice):
                    df = df[(df[ele_] >= slice_.start) & (df[ele_] <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[df[ele_].isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[df[ele_] == slice_]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 1 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()
        headsize = len(fields(Diamond1Head))
        self.head = Diamond1Head(*data[:headsize])
        ncols = len(data[headsize:]) / self.head.nrec
        if not ncols.is_integer():
            raise Exception("Data can't be splitted into integer columns.")
        self.data = np.asarray(data[headsize:]).reshape((self.head.nrec, int(ncols)))
        self._ds = self.to_frame()

    def write(self, pathfile: str, encoding: str = 'GBK',
              fmt: str = '%5i%8.2f%8.2f%5i%5i%5i%5i%5i%5i%5i%5i%5i%5i'
                         '%5i%5i%5i%7.1f%7.1f%5i%7.1f%5i%5i%5i%5i%5i%5i'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str], default: '%.2f'
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head[:3]) + "\n")
            f.write("\t".join(head[3:]) + "\n")

            data = self.to_numpy()
            if len(fmt.split('%')) == data.shape[-1] - 2:
                fmt = '%'.join(fmt.split('%')[:-2])
            np.savetxt(f, data, fmt=fmt)

    @property
    def index_cols(self):
        """base index columns, keep in frame when select data"""
        return ['stid', 'lon', 'lat', 'height', 'level']

    @property
    def dtype(self):
        """column data type"""
        dtypes = [('stid', 'i4'), ('lon', 'f4'), ('lat', 'f4'), ('height', 'f4'), ('level', 'i1'),
                  ('tcc', 'i2'), ('windr', 'i2'), ('winds', 'i2'), ('slp', 'i2'), ('dp3', 'i2'),
                  ('ww1', 'i2'), ('ww2', 'i2'), ('pr6', 'f4'), ('lcs', 'i2'), ('lcc', 'i2'),
                  ('lch', 'i2'), ('dew', 'f4'), ('vis', 'f4'), ('ww', 'i2'), ('t', 'f4'),
                  ('mcs', 'i2'), ('hcs', 'i2'), ('identifier1', 'i2'), ('identifier2', 'i2'),
                  ('dt24', 'f4'), ('dp24', 'f4')]
        if self.data.shape[-1] == len(dtypes) - 2:
            dtypes = dtypes[:-2]
        return np.dtype(dtypes)

    @property
    def names(self):
        """column-name dictionary"""
        names = [('stid', '区站号'), ('lon', '经度'), ('lat', '纬度'), ('height', '拔海高度'),
                 ('level', '站点级别'), ('tcc', '总云量'), ('windr', '风向'), ('winds', '风速'),
                 ('slp', '气压'), ('dp3', '3小时变压'), ('ww1', '过去天气1'), ('ww2', '过去天气2'),
                 ('pr6', '6小时降水'), ('lcs', '低云状'), ('lcc', '低云量'), ('lch', '低云高'),
                 ('dew', '露点'), ('vis', '能见度'), ('ww', '现在天气'), ('t', '温度'),
                 ('mcs', '中云状'), ('hcs', '高云状'),
                 ('identifier1', '标志1'), ('identifier2', '标志2'),
                 ('dt24', '24小时变温'), ('dp24', '24小时变压')]
        return dict(names)

    def to_numpy(self):
        """Convert instance into numpy array with dtypes

        Returns
        -------
        data: np.ndarray
            converted data
        """
        return unstructured_to_structured(self.data, dtype=self.dtype)

    def to_frame(self):
        """Convert instance into pandas.DataFrame

        Returns
        -------
        data: xr.DataArray
            converted data
        """
        data = pd.DataFrame.from_records(self.to_numpy())
        return data

    def __repr__(self):
        """print"""
        return self.head.__repr__()



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2021/2/3 17:26
# @Last Modified by: wqshen

import os
import numpy as np
import xarray as xr
from typing import Optional, Union, cast
from datetime import datetime
from dataclasses import dataclass, fields, astuple


@dataclass
class Diamond11Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    duration: int
    level: int
    xinterval: float
    yinterval: float
    startlon: float
    endlon: float
    startlat: float
    endlat: float
    xsize: int
    ysize: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond11:
    """Micaps Diamond Data class

    From this class, one can read and write Micaps Diamond11 file

    Read Diamond11
    --------------
    >>> m11 = Diamond11(pathfile='../../sample_data/diamond_11/21032920.024')
    >>> m11.read()
    >>> print(m11)
    >>> print(m11.data.shape)
    >>> print(m11.to_xarray())

    Write Diamond11
    --------------
    >>> m11 = Diamond11()
    >>> m11.head = Diamond11Head('diamond', 11, '20年11月27日T63_500hPa风场120小时预报',
    ...                          2020, 11, 27, 20, 120, 500,
    ...                          1.875, -1.875, 0, 180, 90, 0, 97, 49, 20, -300, 300, 1, 0)
    >>> m11.data = np.random.randint(-50, 10, size=(97, 49))
    >>> m11.write("20201127.120", fmt='%.5f')
    """
    head: Optional[Diamond11Head] = None
    data: Optional[np.ndarray] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[Union[xr.DataArray, xr.Dataset]] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, **kwargs):
        """interface to call `xarray.Dataset.sel`_ method

        Parameters
        ----------
        kwargs (dict): other parameters used to process xarray.Dataset variable

        Returns
        -------
        dar (xarray.Dataset): sel dataset
        """
        if self._ds.indexes['lat'].is_monotonic_decreasing:
            if 'lat' in kwargs and isinstance(kwargs.get('lat'), slice):
                lat = kwargs.get('lat')
                kwargs['lat'] = slice(lat.stop, lat.start, lat.step)
        dar = self._ds.sel(**kwargs)
        return dar

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 11 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()

        headsize = len(fields(Diamond11Head))
        self.head = Diamond11Head(*data[:headsize])
        self.data = np.asfarray(data[headsize:]).reshape((2, self.head.ysize, self.head.xsize))
        self._ds = self.to_xarray(name='var')

    def write(self, pathfile: str, encoding: str = 'GBK', fmt: str = '%.2f'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str], default: '.5f'
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head[:3]) + "\n")
            f.write("\t".join(head[3:9]) + "\n")
            f.write("\t".join(head[9:]) + "\n")

            np.savetxt(f, self.data[0], fmt=fmt, delimiter='\t')
            np.savetxt(f, self.data[1], fmt=fmt, delimiter='\t')

    def to_netcdf(self, pathfile: str, name: str, complevel: int = 5):
        """write data into NetCDF4 file

        Parameters
        ----------
        pathfile : str
            path to file name
        name : str
            variable name when pack data into NetCDF4 Variable
        complevel : Optional[int], default: 5
            Compress level when pack data into NetCDF4 dataset
        """
        data = self.to_xarray()
        # convert to dataset and save to netcdf file with encoding
        comp = dict(zlib=True, complevel=complevel)
        ds = data.to_dataset(name=name)
        encoding = {var: comp for var in ds.data_vars}
        ds.to_netcdf(pathfile, encoding=encoding)

    def to_xarray(self, name=('u', 'v')):
        """Convert instance into xarray.DataArray

        Returns
        -------
        data: xr.DataArray
            converted data
        """
        lons = np.linspace(self.head.startlon, self.head.endlon, self.head.xsize)
        lats = np.linspace(self.head.startlat, self.head.endlat, self.head.ysize)
        var_x = xr.DataArray(self.data[0], dims=('lat', 'lon'), coords={'lat': lats, 'lon': lons},
                             name=name[0])
        var_y = xr.DataArray(self.data[1], dims=('lat', 'lon'), coords={'lat': lats, 'lon': lons},
                             name=name[1])

        # To mask potential missing values
        if self.missing_value is not None:
            var_x = var_x.where(var_x == self.missing_value)
            var_y = var_y.where(var_y == self.missing_value)
        # Append extra attributes to netcdf varibale
        var_x.attrs['description'] = self.head.description + "_" + name[0]
        var_y.attrs['description'] = self.head.description + "_" + name[1]
        var_x.attrs['inittime'] = datetime(self.head.year, self.head.month, self.head.day,
                                           self.head.hour)
        var_y.attrs['inittime'] = datetime(self.head.year, self.head.month, self.head.day,
                                           self.head.hour)
        var_x.attrs['fh'] = self.head.duration
        var_y.attrs['fh'] = self.head.duration

        if self.pathfile is not None:
            var_x.attrs['raw'] = self.pathfile
            var_y.attrs['raw'] = self.pathfile
        var = xr.merge([var_x, var_y])
        return var

    def __repr__(self):
        """print"""
        return self.head.__repr__()



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2021/12/17 10:25
# @Last Modified by: wqshen


import os
import numpy as np
import pandas as pd
from typing import Optional, Union, Any, cast
from dataclasses import dataclass, fields, astuple
from numpy.lib.recfunctions import unstructured_to_structured


@dataclass
class Diamond5Head:
    diamond: str
    dtype: int
    description: str
    year: int
    month: int
    day: int
    hour: int
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond5Index:
    stid: str
    lon: float
    lat: float
    height: float
    nrec: int

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            setattr(self, field.name, field.type(field_value))


@dataclass
class Diamond5:
    """Micaps Diamond 5 Data class

    From this class, one can read and write Micaps Diamond 5 file

    Read Diamond5
    --------------
    >>> m5 = Diamond5(pathfile='../../sample_data/diamond_5/18091608.000')
    >>> m5.read()
    >>> print(m5)
    >>> print(m5.data)
    >>> print(m5.to_frame())

    Write Diamond5
    --------------
    >>> m5 = Diamond5()
    >>> m5.head = Diamond5Head('diamond', 5, '18年09月16日08时温度对数压力图', 18, 09, 16, 8, 519)
    >>> m5.data = np.array([
    ...    [67297, 34.90, -19.79, 16, 32, 4, 60, 2, 142, 9999, 0, 0, 0, 32, 4, 300, 22.5, 20.0, 0, 24.2, 20, 10, 0, 0],
    ...    [96315, 114.93, 4.93, 29, 32, 7, 230, 1, 88, 9999, 2, 2, 0, 38, 1, 300, 24.8, 8.0, 5, 26.5, 24, 18]
    ...             ])
    >>> m5.write("18091608.000")
    """
    head: Optional[Diamond5Head] = None
    index: Optional[pd.DataFrame] = None
    data: Optional[list] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[pd.DataFrame] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, element: Any = None, stid: Any = None, lon: Any = None, lat: Any = None,
            height: Any = None, query: str = None):
        """interface to select data

        Parameters
        ----------
        element: Any
            select element in columns
        stid: Any
            select data at given station or station list
        lon: Any
            select data within given longitude range
        lat: Any
            select data within given latitude range
        height: Any
            select data within given height range
        query: str
            pandas.DataFrame.query string to filter data

        Returns
        -------
        df (pd.DataFrame): selected data
        """
        df = self._ds
        if element is not None:
            if isinstance(element, str):
                df = df[[*self.index_cols, element]]
            elif isinstance(element, (list, tuple, np.ndarray, pd.Series)):
                df = df[[*self.index_cols, *element]]

        for slice_, ele_ in zip((stid, lon, lat, height), self.index_cols):
            if slice_ is not None:
                idx = df.index.get_level_values(ele_)
                if isinstance(slice_, slice):
                    df = df[(idx >= slice_.start) & (idx <= slice_.stop)]
                elif isinstance(slice_, (list, tuple, np.ndarray, pd.Series)):
                    df = df[idx.isin(slice_)]
                elif isinstance(slice_, (str, int, float)):
                    df = df[idx == slice_]
                else:
                    raise NotImplementedError(f"{ele_} must be type of int, float, str, list, "
                                              f"tuple, numpy.ndarray, pandas.Series")

        if query is not None:
            df = df.query(query)
        return df

    def read(self, path_or_string: Union[str, bytes], encoding='GBK'):
        """Read micaps diamond 5 file

        Parameters
        ----------
        path_or_string : Union[str, bytes]
            file content str, bytes or path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        """
        if isinstance(path_or_string, bytes):
            data = path_or_string.decode(encoding).split()
        elif os.path.isfile(path_or_string):
            with open(path_or_string, 'r', encoding=encoding) as f:
                data = f.read().split()
        else:
            data = path_or_string.split()

        headsize = len(fields(Diamond5Head))
        self.head = Diamond5Head(*data[:headsize])

        irec, p, indices, records = 0, headsize, [], []
        while irec < self.head.nrec:
            rec_head = data[p:p + 5]
            nrec_irec = int(rec_head[-1])
            p += 5
            record = data[p:p + nrec_irec]
            p += nrec_irec
            irec += 1
            indices.append(rec_head)
            record = np.array(record).reshape(-1, len(self.dtype))
            record = unstructured_to_structured(record, dtype=self.dtype)
            records.append(record)
        index = unstructured_to_structured(np.array(indices), dtype=self.dtype_index)
        self.index = pd.DataFrame.from_records(index)
        self.data = records
        self._ds = self.to_frame()

    def write(self, pathfile: str, encoding: str = 'GBK',
              fmt='%6d %6d %4d %4d %4d %4d'):
        """Write data into micaps diamond4 text file

        Parameters
        ----------
        pathfile : str
            path to file name
        encoding : str, Optional, default: 'GBK'
            encoding of output file
        fmt : Optional[str], default: '%.2f'
            float format to save data
        """
        with open(pathfile, 'w', encoding=encoding) as f:
            head = tuple(map(str, astuple(self.head)))
            f.write("\t".join(head[:3]) + "\n" + "\t".join(head[3:]) + '\n')
            for i, (head, track) in enumerate(zip(self.index.values, self.data)):
                f.write("{:6s}\t{:.2f}\t{:.2f}\t{:4.0f}\t{:4d}".format(*head) + "\n")
                np.savetxt(f, track, fmt=fmt)

    @property
    def dtype(self):
        dtypes = [('pressure', 'f4'), ('height', 'f4'), ('temperature', 'f4'),
                  ('dewpoint', 'f4'), ('windr', 'f4'), ('winds', 'f4')]
        return np.dtype(dtypes)

    @property
    def dtype_index(self):
        dtypes = [('stid', 'U6'), ('lon', 'f4'), ('lat', 'f4'), ('height', 'f4'), ('nrec', 'i4')]
        return np.dtype(dtypes)

    @property
    def index_cols(self):
        """base index columns, keep in frame when select data"""
        return ['stid', 'lon', 'lat', 'height']

    def to_numpy(self):
        """Convert instance into numpy array with dtypes

        Returns
        -------
        data: list
            converted data
        """
        return [unstructured_to_structured(np.asarray(d).reshape(-1, len(self.dtype)),
                                           dtype=self.dtype)
                for d in self.data]

    def to_frame(self):
        """Convert instance into pandas.DataFrame

        Returns
        -------
        data: xr.DataFrame
            converted data
        """
        nrecs = self.index.nrec // len(self.dtype)
        multi_index = self.index.loc[self.index.index.repeat(nrecs)]
        multi_index = pd.MultiIndex.from_frame(multi_index[self.index_cols])
        data = pd.DataFrame.from_records(np.hstack(self.data), index=multi_index)
        return data

    def __repr__(self):
        """print"""
        return self.head.__repr__()

    def __getitem__(self, stid: str):
        """get data of station"""
        if isinstance(stid, str):
            return self._ds[self._ds.stid == stid]
        elif isinstance(stid, (tuple, list, np.ndarray)):
            return self._ds[self._ds.stid.isin(stid)]
        else:
            TypeError("stid must be type of str")



# -*- coding: utf-8 -*-
# @Author: wqshen
# @Email: wqshen91@gmail.com
# @Date: 2023/2/9 22:07
# @Last Modified by: wqshen

import numpy as np
import xarray as xr
from typing import Optional, Union, cast
from struct import unpack, calcsize
from datetime import datetime, timedelta
from dataclasses import dataclass, fields


@dataclass
class LatLonHead:
    dataname: str
    varname: str
    units: str
    label: int
    unitlen: int
    slat: float
    wlon: float
    nlat: float
    elon: float
    clat: float
    clon: float
    rows: int
    cols: int
    dlat: float
    dlon: float
    nodata: float
    levelbytes: int
    levelnum: int
    amp: int
    compmode: int
    dates: int
    seconds: int
    min_value: int
    max_value: int
    reserved: list

    @property
    def dtype(self) -> str:
        return '=128s32s16sHhffffffiifffihhhHihh6h'

    def __post_init__(self):
        self._cast_fields_types()

    def _cast_fields_types(self):
        for field in fields(cast(dataclass, self)):
            field_value = getattr(self, field.name)
            if isinstance(field_value, bytes):
                field_value = field_value.rstrip(b'\x00\xfe').decode('utf8')
            setattr(self, field.name, field.type(field_value))


@dataclass
class LatLon(object):
    head: Optional[LatLonHead] = None
    data: Optional[Union[np.ndarray, dict]] = None
    pathfile: Optional[str] = None
    missing_value: Optional[float] = None
    _ds: Optional[xr.DataArray] = None

    def __post_init__(self):
        if self.pathfile is not None:
            self.read(self.pathfile)

    def sel(self, **kwargs):
        """interface to call `xarray.DataArray.sel`_ method

        Parameters
        ----------
        kwargs (dict): other parameters used to process xarray.DataArray variable

        Returns
        -------
        dar (xarray.DataArray): sel variable
        """
        if self._ds.indexes['lat'].is_monotonic_decreasing:
            if 'lat' in kwargs and isinstance(kwargs.get('lat'), slice):
                lat = kwargs.get('lat')
                kwargs['lat'] = slice(lat.stop, lat.start, lat.step)
        dar = self._ds.sel(**kwargs)
        return dar

    def read(self, path_or_bytes: Union[str, bytes]):
        """read LatLon file or bytes

        Parameters
        ----------
        path_or_bytes: path to latlon file or data in bytes
        """
        if isinstance(path_or_bytes, str):
            with open(path_or_bytes, 'rb') as f:
                bytes_array = f.read()
        else:
            bytes_array = path_or_bytes
            print(type(path_or_bytes))
        head_type = '=128s32s16sHhffffffiifffihhhHihh6h'
        p = calcsize(head_type)
        head_data = unpack(head_type, bytes_array[:p])
        self.head = LatLonHead(*head_data[:-6], list(head_data[-6:]))

        body_type = f'{self.head.levelbytes // 2}h'
        body_data = np.frombuffer(bytes_array[p:p + calcsize(body_type)], dtype=body_type)[0]
        if self.head.compmode != 0:
            data = self.decompress(body_data)
        else:
            data = np.array(body_data).reshape(self.head.rows, self.head.cols)
        data[data >= self.head.min_value] *= 1.0 / self.head.amp
        data[data < 0] = np.nan
        self.data = data
        self._ds = self.to_xarray()

    def decompress(self, buf: np.ndarray) -> np.ndarray:
        """decompress data

        Parameters
        ----------
        buf: np.ndarray
            compressed data

        Returns
        -------
        data: np.ndarray
            decompressed data
        """
        head = self.head
        data = np.ones((head.rows, head.cols))
        data *= int(head.nodata * head.amp + 0.5)
        y, x, n = buf[:3]
        buf = buf[3:]
        while y >= 0 and x >= 0 and n > 0:
            data[y, x:x + n] = buf[:n]
            buf = buf[n:]
            y, x, n = buf[:3]
            buf = buf[3:]
        return data

    def to_xarray(self) -> xr.DataArray:
        head = self.head
        lons = np.linspace(head.wlon, head.elon - head.dlon, head.cols)
        lats = np.linspace(head.slat + head.dlat, head.nlat, head.rows)
        time = datetime(1970, 1, 1) + timedelta(days=int(head.dates) - 1, seconds=int(head.seconds))
        attrs = {
            k: getattr(head, k)
            for k in ('dataname', 'varname', 'units', 'label',
                      'unitlen', 'dates', 'seconds', 'levelnum')
        }
        if self.pathfile is not None:
            attrs['raw'] = self.pathfile

        name = head.varname
        dims = ('time', 'lat', 'lon')
        coords = {'time': [time], 'lat': lats, 'lon': lons}
        data = xr.DataArray(self.data[None, ...], dims=dims, coords=coords, name=name, attrs=attrs)

        return data

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close open dataset"""
        del self.pathfile, self.data, self.head, self._ds

    def __repr__(self) -> str:
        """print"""
        return self.head.__repr__()



#!/usr/bin/env python
# coding=utf8
# Many method of MdfsClient rewritten from Sean on http://www.micaps.cn/MifunForum/post/list?tId=123
# Rewritten by Wenqiang Shen at 2018/8/9 Zhejiang
# Email: wqshen91@gmail.com

import os
import re
import inspect
import importlib
from struct import unpack
from httplib2 import Http
from logzero import logger
from retrying import retry
from datetime import datetime
from itertools import product
from urllib.parse import urljoin
from google.protobuf.message import DecodeError
from .mdfs import DataBlock_pb2 as DataBlock


class MdfsClient(Http):
    def __init__(self, address):
        """GDS Data Client

        Parameters
        ----------
        address: IP address and port,  xxx.xxx.xxx.xxx:xxxx

        Notes
        -----
        You can find ip and port of GDS Server in micaps.exe.config/appSetting/GDSServer1
        """
        super(MdfsClient, self).__init__()
        self.__basicUrl = f"http://{address}/DataService"

    def sel(self, datasource, inittime, fh=None, varname=None, **kwargs):
        """interface to select variable from file by given more filter and clip parameters

        Parameters
        ----------
        datasource (str): data source name, see Registry/DATA
        inittime (datetime): model initial datetime or observation datetime
        fh (int): forecast hour
        varname (str): variable name
        timeout (int): timeout to collect data
        kwargs (dict): other k/v arguments passed to `sel` method of specific reader

        Returns
        -------
        (xarray.DataArray, list[xarray.DataArray]): Readed variable in xarray.DataArray
        """
        datasource = [datasource] if isinstance(datasource, str) else datasource
        inittime = [inittime] if isinstance(inittime, datetime) else inittime
        fh = [fh] if isinstance(fh, int) or fh is None else fh
        varname = [varname] if isinstance(varname, str) or varname is None else varname
        requests = list(product(datasource, inittime, fh, varname))
        datas = [None] * len(requests)

        for i, request in enumerate(requests):
            logger.debug("{}".format(requests))
            try:
                data = self._sel(*request, **kwargs)
            except Exception as e:
                logger.exception("Worker: {} - {}".format(request, e))
                continue
            datas[i] = data
        if all(v is None for v in datas):
            raise Exception("All requests failed.")
        return datas if len(requests) > 1 else datas[0]

    def _sel(self, datasource, inittime=None, fh=None, varname=None, **kwargs):
        """Sel file variable on remote workers

        Parameters
        ----------
        inittime (datetime): model initial datetime
        datasource (str): data source name, see Registry/DATA
        fh (int): forecast hour
        varname (str): variable name
        wildcard (str): file name wildcard, speedup runtime if offered,
        **kwargs: other k/v arguments passed to `sel` method of specific reader

        Returns
        -------
        (xarray.DataArray): variable
        """
        directory = f'{datasource}/{varname}'
        level = kwargs.pop('level', None)
        if level is not None:
            directory += f'/{level}'

        if fh is None:
            fh = 0

        wildcard = kwargs.pop('wildcard', None)
        if wildcard is None:
            wildcard = self.guess_filename_wildcard(directory)

        filename = wildcard.format(inittime=inittime, fh=fh)
        ret = self.get_data(directory, filename)
        cls = self.guess_interface_class(filename, ret.byteArray)
        cls.read(ret.byteArray)
        return cls.sel(**kwargs)

    def guess_interface_class(self, filename: str, contents: bytes):
        """

        Parameters
        ----------
        filename: str
            file name
        contents: bytes
            GDS getData

        Returns
        -------
        cls: class
            interface class
        """
        if filename.upper().endswith('AWX'):
            from awx import Awx
            cls = Awx()
        elif filename.upper().endswith('LATLON'):
            from .latlon import LatLon
            cls = LatLon()
        else:
            discriminator, dtype = unpack('4sh', contents[:6])
            if discriminator == b'mdfs':
                if dtype in (4, 11):
                    from .mdfs.mdfs_grid_data import MdfsGridData
                    cls = MdfsGridData()
                else:
                    from .mdfs.mdfs_station_data import MdfsStationData
                    cls = MdfsStationData()
            else:
                discriminator, dtype = contents.decode().split()[:2]
                if discriminator.lower() == 'diamond':
                    Cls = getattr(importlib.import_module(f'.diamond{dtype}',
                                                          'pymdfs.diamond'),
                                  f'Diamond{dtype}')
                    cls = Cls()
                else:
                    raise NotImplementedError(f"Unsupported file format.")
        if cls is None:
            raise Exception("Failed to guess file type.")
        return cls

    def guess_filename_wildcard(self, directory: str) -> str:
        filelist = self.get_file_list(directory)
        latest = sorted(list(filelist.resultMap.keys()))[-1]
        if re.match(r'.*(\d{14})\.(\d{3})', latest):
            wildcard = '{inittime:%Y%m%d%H%M%S}.{fh:03d}'
        elif re.match(r'.*(\d{8})\.(\d{3})$', latest):
            wildcard = '{inittime:%y%m%d%H}.{fh:03d}'
        elif re.match(r'.*(\d{8})([._])(\d{6}).*', latest):
            wildcard = re.sub(r'(.*)(\d{8})([._])(\d{6})(.*)',
                              r'\g<1>{inittime:%Y%m%d}\g<3>{inittime:%H%M%S}\g<5>',
                              latest)
        elif re.match(r'.*(\d{8})([._])(\d{4}).*', latest):
            wildcard = re.sub(r'(.*)(\d{8})([._])(\d{4})(.*)',
                              r'\g<1>{inittime:%Y%m%d}\g<3>{inittime:%H%M}\g<5>',
                              latest)
        elif re.match(r'.*(\d{14}).*', latest):
            wildcard = re.sub(r'(.*)(\d{14})(.*)', r'\g<1>{inittime:%Y%m%d%H%M%S}\g<3>', latest)
        else:
            raise NotImplementedError(f"filename {latest} can not be guess wildcard.")

        return wildcard

    def get_latest_data_name(self, directory: str, filter: str = "*") -> DataBlock.StringResult:
        """get latest data filename in GDS server at given directory and wildcard

        Parameters
        ----------
        directory: str
            data path in GDS server, like "ECMWF_HR/TMP/850"
        filter: str
            filename wildcard, for instance, "*.024" represents file name ends with 024

        Returns
        -------
        string_result: DataBlock.StringResult
        """
        url = self.get_concate_url("getLatestDataName", directory, "", filter, "")
        response, content = self.request(url)
        string_result = DataBlock.StringResult()
        string_result.ParseFromString(content)
        self.result_check(string_result)
        return string_result

    @retry(stop_max_attempt_number=3)
    def get_file_list(self, directory: str) -> DataBlock.MapResult:
        """get file list in GDS server at given directory

        Parameters
        ----------
        directory: str
            data path in GDS server, like "ECMWF_HR/TMP/850"

        Returns
        -------
        map_result: DataBlock.MapResult
        """
        url = self.get_concate_url("getFileList", directory, "", "", "")
        response, content = self.request(url)
        map_result = DataBlock.MapResult()
        map_result.ParseFromString(content)
        self.result_check(map_result)
        return map_result

    def list_vars(self, directory: str) -> DataBlock.MapResult:
        """list variables in GDS server at given directory

        Parameters
        ----------
        directory: str
            data path in GDS server, like "ECMWF_HR"

        Returns
        -------
        map_result: DataBlock.MapResult
        """
        result = self.get_file_list(directory)
        variables = [k for k, v in result.resultMap.items() if v == 'D']
        return variables

    @retry(stop_max_attempt_number=3)
    def get_data(self, directory: str, filename: str) -> DataBlock.ByteArrayResult:
        """download specific file from GDS server

        Parameters
        ----------
        directory: str
            data path in GDS server, like "ECMWF_HR/TMP/850"
        filename: str
            filename in directory

        Returns
        -------
        byte_array_result: DataBlock.ByteArrayResult
        """
        url = self.get_concate_url("getData", directory, filename, "", "")
        logger.debug(url)
        response, content = self.request(url)
        byte_array_result = DataBlock.ByteArrayResult()
        byte_array_result.ParseFromString(content)
        self.result_check(byte_array_result)
        return byte_array_result

    def write_byte_array(self, url: str, byte_array: str):
        """write contents to GDS server

        Parameters
        ----------
        url: str
            url in GDS server
        byte_array: str
        """
        new_url = self.get_concate_url("writeByteArrayDataWithURL", "", "", "", url)
        return self.send_http_message(new_url, byte_array)

    def send_http_message(self, url: str, byte_array):
        """POST content to server

        Parameters
        ----------
        url: str
        byte_array
        """
        # 异常情况,不应出现,健壮性保护
        if byte_array is None or len(byte_array) == 0 or url is None or len(url) == 0:
            raise Exception("Argument Error.")
        start_time = datetime.now()
        self.request(url, method='POST', headers={"content-type": "application/octet-stream"})
        pass

    def get_concate_url(self, requestType: str, directory: str = None, fileName: str = None,
                        filter: str = None, url: str = None):
        """concate all arguments to GDS url

        Parameters
        ----------
        requestType: str
        directory: str
            data path in GDS server, like "ECMWF_HR/TMP/850"
        fileName: str
            filename in directory
        filter: str
            filename wildcard, for instance, "*.024" represents file name ends with 024
        url: str
            url in GDS server for write bytes array

        Returns
        -------
        url: str
            concate url
        """
        new_url = [self.__basicUrl]
        new_url.append("?requestType=" + requestType)  # requestType始终不为空
        if (directory is not None) and len(directory) > 0:
            new_url.append("&directory=" + directory)
        if (fileName is not None) and len(fileName) > 0:
            new_url.append("&fileName=" + fileName)
        if (filter is not None) and len(filter) > 0:
            new_url.append("&filter=" + filter)
        if (url is not None) and len(url) > 0:
            new_url.append("&url=" + url)
        url = "".join(new_url)
        return url

    def result_check(self, result):
        if result is None:  # 读取服务器发生IO异常
            raise Exception("{}：读取GDS服务器发生IO异常,程序结束".format(inspect.stack()[1][3]))
        if result.errorCode != 0:  # 没有发生IO异常, 但出现其它错误, 比如数据不存在等
            if result.errorMessage == "NotFoundException":
                raise FileNotFoundError("Error, file not found.")
            raise Exception(
                "{}: Code {}, Message {}, 程序结束。".format(inspect.stack()[1][3], result.errorCode,
                                                            result.errorMessage))

    def walk(self, directory: str):
        """recursive get all file path of given path in GDS server to setup a files tree

        Parameters
        ----------
        directory: str
            path to GDS server
        """
        if directory[-1] != "/":
            directory += '/'
        result = self.get_file_list(directory)
        dirs, nondirs, nondirsize, walk_dirs = [], [], [], []
        dirs.extend([k for k, v in result.resultMap.items() if v == 'D'])
        nondirs.extend([k for k, v in result.resultMap.items() if v != 'D'])
        nondirsize.extend([int(v) for k, v in result.resultMap.items() if v != 'D'])
        for dirname in dirs:
            new_path = urljoin(directory, dirname)
            yield from self.walk(new_path)
        yield directory, dirs, nondirs, nondirs

    def get_path_file_list(self, path: str):
        """recursive get all file path of given path in GDS server

        Parameters
        ----------
        path: str
            path to GDS server
        """
        for top, dir, files, sizes in self.walk(path):
            if len(dir) == 0 and len(files) > 0:
                for file in files:
                    yield os.path.join(top, file)

    def download(self, pathfile: str, filesize: int = None, outdir: str = "S:/micaps") -> int:
        """Download file from GDS server to local disk

        Parameters
        ----------
        pathfile: str
            path to file in GDS server
        filesize: int
            file size
        outdir: str
            path to save downloaded files

        Returns
        -------
         0 - file has existed
        -1 - Decode Error
        -2 - File not found
        """

        logger.debug("Download file {} and save to {}".format(pathfile, outdir))
        directory = os.path.dirname(pathfile)
        filename = os.path.basename(pathfile)
        outpath = "{}/{}".format(outdir, directory)
        if not os.path.exists(outpath):
            os.makedirs(outpath)
        pathfile = os.path.join(outpath, filename)
        if os.path.isfile(pathfile):
            logger.warning('{}文件已存在，跳过。'.format(pathfile))
            return 0
        try:
            result = self.get_data(directory, filename)
        except DecodeError:
            logger.error('google.protobuf.message.DecodeError! Skip file {}'.format(pathfile))
            return -1
        except FileNotFoundError:
            logger.error('File not found, skip file {}'.format(pathfile))
            return -2
        with open(pathfile, "wb") as f:
            if filesize is not None:
                logger.debug("{} - {:.1f}kb - {:.1f}kb".format(filename, filesize / 1024.,
                                                               len(result.byteArray) / 1024.))
            else:
                logger.debug("{} - {:.1f}kb".format(filename, len(result.byteArray) / 1024.))
            f.write(result.byteArray)
        if filesize is not None:
            self.file_size_verify(pathfile, filesize)

    def file_size_verify(self, pathfile: str, size: int) -> int:
        """verify file size

        Parameters
        ----------
        pathfile: str
            path to local file
        size: int
            expected file size
        Returns
        -------
        -1 for failure, 0 for success
        """
        fkbytes = os.path.getsize(pathfile) / 1024
        wkbytes = size / 1024
        if wkbytes - fkbytes > 1:
            logger.warning("{}：File {} size verify failed!".format(inspect.stack()[1][3], pathfile))
            return -1
        else:
            return 0

    def parallel_download(self, filelist: list, outdir: str = "S:/micaps", njobs: int = 4):
        """parallel download file in GDS server

        Parameters
        ----------
        filelist: list
            file list to download
        outdir: str
            path to save downloaded files
        njobs: int
            number of cpu counts in parallel download
        """
        from pathos.multiprocessing import ProcessingPool

        pool = ProcessingPool(njobs)
        f = lambda file: self.download(file, outdir=outdir)
        pool.map(f, filelist)
        pool.close()
        pool.join()

    def directory_sync(self, directory: str, outdir: str = "S:/micaps", njobs: int = 4):
        """sync local and gds server

        Parameters
        ----------
        directory: str
            data path in GDS Server
        outdir: str
            local path
        njobs: int
            number of cpu counts in parallel download
        """
        pathfiles = self.get_path_file_list(directory)
        self.parallel_download(pathfiles, outdir=outdir, njobs=njobs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.__basicUrl

